/*--------------------------------------------------
   TGB Dual - Gameboy Emulator -
   Copyright (C) 2001  Hii

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

//--------------------------------------------------
// GB 饹,¾

#include <stdio.h>
#include <list>

#include "gb_types.h"
#include "renderer.h"

#define INT_VBLANK 1
#define INT_LCDC 2
#define INT_TIMER 4
#define INT_SERIAL 8
#define INT_PAD 16

class gb;
class cpu;
class lcd;
class apu;
class apu_snd;
class rom;
class mbc;
class cheat;

struct ext_hook{
	byte (*send)(byte);
	bool (*led)(void);
};

struct cheat_dat{
	bool enable;
	byte code;
	word adr;
	byte dat;
	char name[255];
	cheat_dat *next;
};

struct gb_regs {
	byte P1,SB,SC,DIV,TIMA,TMA,TAC,IF,LCDC,STAT,SCY,SCX,LY,LYC,DMA,BGP,OBP1,OBP2,WY,WX,IE;
};

struct gbc_regs {
	byte KEY1,VBK,HDMA1,HDMA2,HDMA3,HDMA4,HDMA5,RP,BCPS,BCPD,OCPS,OCPD,SVBK;
};

union pare_reg {
	word w;
	struct{byte l,h;}b;
};

struct cpu_regs {
	pare_reg AF;
	pare_reg BC;
	pare_reg DE;
	pare_reg HL;
	word SP;
	word PC;
	byte I;
};

struct apu_stat{
	bool sq1_playing;
	int sq1_sw_time;
	int sq1_sw_dir;
	int sq1_sw_shift;

	int sq1_len;
	int sq1_init_len;
	int sq1_type;

	int sq1_vol;
	int sq1_init_vol;
	int sq1_env_dir;
	int sq1_env_speed;

	int sq1_freq;
	int sq1_init_freq;

	int sq1_hold;


	bool sq2_playing;

	int sq2_len;
	int sq2_init_len;
	int sq2_type;

	int sq2_vol;
	int sq2_init_vol;
	int sq2_env_dir;
	int sq2_env_speed;

	int sq2_freq;
	int sq2_init_freq;

	int sq2_hold;


	bool wav_playing;
	int wav_vol;
	int wav_freq;
	int wav_init_freq;
	int wav_init_len;
	int wav_len;
	int wav_hold;


	bool noi_playing;
	int noi_len;
	int noi_init_len;

	int noi_vol;
	int noi_init_vol;
	int noi_env_dir;
	int noi_env_speed;

	int noi_freq;
	int noi_init_freq;
	int noi_hold;
	int noi_step;

	int master_enable;
	int ch_enable[4][2];
	int master_vol[2];
	int ch_on[4];
	int wav_enable;
};

struct apu_que {
	word adr;
	byte dat;
	int clock;
};

struct rom_info {
	char cart_name[18];
	int cart_type;
	byte rom_size;
	byte ram_size;

	bool check_sum;
	int gb_type;
};

class gb
{
friend class cpu;
public:
	gb(renderer *ref,bool b_lcd,bool b_apu);
	~gb();

	cpu *get_cpu() { return m_cpu; }
	lcd *get_lcd() { return m_lcd; }
	apu *get_apu() { return m_apu; }
	rom *get_rom() { return m_rom; }
	mbc *get_mbc() { return m_mbc; }
	renderer *get_renderer() { return m_renderer; }
	cheat *get_cheat() { return m_cheat; }
	gb *get_target() { return target; }
	gb_regs *get_regs() { return &regs; }
	gbc_regs *get_cregs() { return &c_regs; }

	void run();
	void reset();
	void set_skip(int frame);
	void set_use_gba(bool use) { use_gba=use; }
	bool load_rom(byte *buf,int size,byte *ram,int ram_size);
	void save_state(FILE *file);
	void restore_state(FILE *file);

	void refresh_pal();

	void set_target(gb *tar) { target=tar; }

	void hook_extport(ext_hook *ext);
	void unhook_extport();

private:
	cpu *m_cpu;
	lcd *m_lcd;
	apu *m_apu;
	rom *m_rom;
	mbc *m_mbc;
	renderer *m_renderer;

	cheat *m_cheat;

	gb *target;

	gb_regs regs;
	gbc_regs c_regs;

	word dmy[160*5]; // vframe Ϥ߽Ф
	word vframe[160*(144+100)];

	ext_hook hook_proc;

	int skip,skip_buf;
	int now_frame;
	int re_render;

	bool hook_ext;
	bool use_gba;
};

class cheat
{
public:
	cheat(gb *ref);
	~cheat();

	byte cheat_read(word adr);
	void cheat_write(word adr,byte dat);

	bool cheak_cheat(word adr);
	void create_cheat_map();

	void add_cheat(cheat_dat *dat);
	void delete_cheat(char *name);
	std::list<cheat_dat>::iterator find_cheat(char *name);
	void create_unique_name(char *buf);

	void clear();

	std::list<cheat_dat>::iterator get_first() { return cheat_list.begin(); }
	std::list<cheat_dat>::iterator get_end() { return cheat_list.end(); }

	int *get_cheat_map() { return cheat_map; }

	void save(FILE *file);
	void load(FILE *file);

private:
	std::list<cheat_dat> cheat_list;
	int cheat_map[0x10000];

	gb *ref_gb;
};

class lcd
{
public:
	lcd(gb *ref);
	~lcd();

	void render(void *buf,int scanline);
	void reset();
	void clear_win_count() { now_win_line=9; }
	word *get_pal(int num) { return col_pal[num]; }
	word *get_mapped_pal(int num) { return mapped_pal[num]; }

	void set_enable(int layer,bool enable);
	bool get_enable(int layer);

	int get_sprite_count() { return sprite_count; };

private:
	void bg_render(void *buf,int scanline);
	void win_render(void *buf,int scanline);
	void sprite_render(void *buf,int scanline);
	void bg_render_color(void *buf,int scanline);
	void win_render_color(void *buf,int scanline);
	void sprite_render_color(void *buf,int scanline);

	word m_pal16[4];
	dword m_pal32[4];
	word col_pal[16][4];
	word mapped_pal[16][4];

	int trans_count;
	byte trans_tbl[160+160],priority_tbl[320];

	int now_win_line;
	int mul;
	int sprite_count;

	bool layer_enable[3];

	gb *ref_gb;
};

class apu
{
friend class apu_snd;
public:
	apu(gb *ref);
	~apu();

	apu_snd *get_renderer() { return snd; }
	apu_stat *get_stat();
	apu_stat *get_stat_cpy();
	byte *get_mem();

	byte read(word adr);
	void write(word adr,byte dat,int clock);

	void update();
	void reset();

private:
	gb *ref_gb;
	apu_snd *snd;
};

class apu_snd : public sound_renderer
{
friend class apu;
public:
	apu_snd(apu *papu);
	virtual ~apu_snd();

	void set_enable(int ch,bool enable);
	bool get_enable(int ch);
	void set_echo(bool echo){ b_echo=echo; };
	void set_lowpass(bool lowpass){ b_lowpass=lowpass; };
	bool get_echo(){ return b_echo; };
	bool get_lowpass(){ return b_lowpass; };


	void render(short *buf,int sample);
	void reset();

private:
	void process(word adr,byte dat);
	void update();
	short sq1_produce(int freq);
	short sq2_produce(int freq);
	short wav_produce(int freq,bool interpolation);
	short noi_produce(int freq);

	apu_stat stat;
	apu_stat stat_cpy,stat_tmp;
	apu_que write_que[0x10000];
	int que_count;
	int bef_clock;
	apu *ref_apu;

	bool b_echo;
	bool b_lowpass;

	byte mem[0x100];
	bool b_enable[4];
};

class mbc
{
public:
	mbc(gb *ref);
	~mbc();

	byte *get_rom() { return rom_page; }
	byte *get_sram() { return sram_page; }
	bool is_ext_ram() { return ext_is_ram; }
	void set_ext_is(bool ext) { ext_is_ram=ext; }

	int get_state();
	void set_state(int dat);
	void set_page(int rom,int sram);

	byte mbc::read(word adr);
	void mbc::write(word adr,byte dat);
	byte mbc::ext_read(word adr);
	void mbc::ext_write(word adr,byte dat);
	void reset();

private:
	void mbc1_write(word adr,byte dat);
	void mbc2_write(word adr,byte dat);
	void mbc3_write(word adr,byte dat);
	void mbc5_write(word adr,byte dat);
	void mbc7_write(word adr,byte dat);
	void huc1_write(word adr,byte dat);
	void huc3_write(word adr,byte dat);
	void tama5_write(word adr,byte dat);
	void mmm01_write(word adr,byte dat);

	byte *rom_page;
	byte *sram_page;

	bool mbc1_16_8;
	byte mbc1_dat;

	byte mbc3_latch; // 1 bits
	byte mbc3_sec; // 6
	byte mbc3_min; // 6
	byte mbc3_hour; // 5
	byte mbc3_dayl; // 8
	byte mbc3_dayh; // 1

	byte mbc3_timer; // 4
	bool ext_is_ram; // 1
	// total 32bits

	int mbc5_dat;

	bool mbc7_write_enable;
	bool mbc7_idle;
	byte mbc7_cs;
	byte mbc7_sk;
	byte mbc7_op_code;
	byte mbc7_adr;
	word mbc7_dat;
	byte mbc7_ret;
	byte mbc7_state;
	word mbc7_buf;
	byte mbc7_count;

	bool huc1_16_8;
	byte huc1_dat;

	gb *ref_gb;
};

class rom
{
public:
	rom();
	~rom();

	rom_info *get_info() { return &info; }
	byte *get_rom() { return first_page; }
	byte *get_sram() { return sram; }
	bool get_loaded() { return b_loaded; }

	void set_first(int page) { first_page=dat+0x4000*page; }

	bool load_rom(byte *buf,int size,byte *ram,int ram_size);

private:
	gb *ref_gb;
	rom_info info;

	byte *dat;
	byte *sram;

	byte *first_page;

	bool b_loaded;
};

class cpu
{
friend class gb;
public:
	cpu(gb *ref);
	~cpu();

	byte read(word adr) { return (ref_gb->get_cheat()->get_cheat_map()[adr])?ref_gb->get_cheat()->cheat_read(adr):read_direct(adr); }

	inline byte read_direct(word adr);
	inline void write(word adr,byte dat);
	inline word readw(word adr) { return read(adr)|(read(adr+1)<<8); }
	inline void writew(word adr,word dat) { write(adr,(byte)dat);write(adr+1,dat>>8); }

	void exec(int clocks);
	byte seri_send(byte dat);
	void irq(int irq_type);
	inline void irq_process();
	void reset();
	void set_trace(bool trace) { b_trace=trace; }

	byte *get_vram() { return vram; }
	byte *get_ram() { return ram; }
	byte *get_oam() { return oam; }
	byte *get_stack() { return stack; }

	byte *get_ram_bank() { return ram_bank; }
	void set_ram_bank(int bank) { ram_bank=ram+bank*0x1000; }

	cpu_regs *get_regs() { return &regs; }

	int get_clock() { return total_clock; }
	bool get_speed() { return speed; }

	bool *get_halt() { return &halt; }

	void save_state(int *dat);
	void restore_state(int *dat);
	void save_state_ex(int *dat);
	void restore_state_ex(int *dat);

private:
	inline byte io_read(word adr);
	inline void io_write(word adr,byte dat);
	byte op_read() { return read(regs.PC++); }
	word op_readw() { regs.PC+=2;return readw(regs.PC-2); }

	int dasm(char *S,byte *A);
	void log();

	gb *ref_gb;
	cpu_regs regs;

	byte ram[0x2000*4];
	byte vram[0x2000*2];
	byte stack[0x80];
	byte oam[0xA0];
	byte spare_oam[0x18];
	byte ext_mem[16];

	byte *vram_bank;
	byte *ram_bank;

	byte z802gb[256],gb2z80[256];
	dword rp_que[256];
	int que_cur;
//	word org_pal[16][4];
	int total_clock,rest_clock,sys_clock,seri_occer,div_clock;
	bool halt,speed,speed_change,dma_executing;
	bool b_trace;
	int dma_src;
	int dma_dest;
	int dma_rest;
	int gdma_rest;
	bool b_dma_first;

	int last_int;
	bool int_desable;

	byte *dma_src_bank;
	byte *dma_dest_bank;

	byte _ff6c,_ff72,_ff73,_ff74,_ff75;

//	FILE *file;
};

inline byte cpu::read_direct(word adr)
{
	switch(adr>>13){
	case 0:
	case 1:
		return ref_gb->get_rom()->get_rom()[adr];//ROMΰ
	case 2:
	case 3:
		return ref_gb->get_mbc()->get_rom()[adr];//Х󥯲ǽROM
	case 4:
		return vram_bank[adr&0x1FFF];//8KBVRAM
	case 5:
		if (ref_gb->get_mbc()->is_ext_ram())
			return ref_gb->get_mbc()->get_sram()[adr&0x1FFF];//ȥåRAM
		else
			return ref_gb->get_mbc()->ext_read(adr);
	case 6:
		if (adr&0x1000)
			return ram_bank[adr&0x0fff];
		else
			return ram[adr&0x0fff];
	case 7:
		if (adr<0xFE00){
			if (adr&0x1000)
				return ram_bank[adr&0x0fff];
			else
				return ram[adr&0x0fff];
		}
		else if (adr<0xFEA0)
			return oam[adr-0xFE00];//object attribute memory
		else if (adr<0xFF00)
			return spare_oam[(((adr-0xFFA0)>>5)<<3)|(adr&7)];
		else if (adr<0xFF80)
			return io_read(adr);//I/O
		else if (adr<0xFFFF)
			return stack[adr-0xFF80];//stack
		else
			return io_read(adr);//I/O
	}
	return 0;
}

void cpu::write(word adr,byte dat)
{
	switch(adr>>13){
	case 0:
	case 1:
	case 2:
	case 3:
		ref_gb->get_mbc()->write(adr,dat);
		break;
	case 4:
		vram_bank[adr&0x1FFF]=dat;
		break;
	case 5:
		if (ref_gb->get_mbc()->is_ext_ram())
			ref_gb->get_mbc()->get_sram()[adr&0x1FFF]=dat;//ȥåRAM
		else
			ref_gb->get_mbc()->ext_write(adr,dat);
		break;
	case 6:
		if (adr&0x1000)
			ram_bank[adr&0x0fff]=dat;
		else
			ram[adr&0x0fff]=dat;
		break;
	case 7:
		if (adr<0xFE00){
			if (adr&0x1000)
				ram_bank[adr&0x0fff]=dat;
			else
				ram[adr&0x0fff]=dat;
		}
		else if (adr<0xFEA0)
			oam[adr-0xFE00]=dat;
		else if (adr<0xFF00)
			spare_oam[(((adr-0xFFA0)>>5)<<3)|(adr&7)]=dat;
		else if (adr<0xFF80)
			io_write(adr,dat);//I/O
		else if (adr<0xFFFF)
			stack[adr-0xFF80]=dat;//stack
		else
			io_write(adr,dat);//I/O
		break;
	}
}

byte cpu::io_read(word adr)
{
	byte ret;
	switch(adr){
	case 0xFF00://P1(ѥå)
		int tmp;
		tmp=ref_gb->get_renderer()->check_pad();
		if (ref_gb->get_regs()->P1==0x03)
			return 0xff;
		switch((ref_gb->get_regs()->P1>>4)&0x3){
		case 0:
			return 0xC0|((tmp&0x81?0:1)|(tmp&0x42?0:2)|(tmp&0x24?0:4)|(tmp&0x18?0:8));
		case 1:
			return 0xD0|((tmp&0x01?0:1)|(tmp&0x02?0:2)|(tmp&0x04?0:4)|(tmp&0x08?0:8));
		case 2:
			return 0xE0|((tmp&0x80?0:1)|(tmp&0x40?0:2)|(tmp&0x20?0:4)|(tmp&0x10?0:8));
		case 3:
			return 0xFF;
		}
		return 0x00;
	case 0xFF01://SB(ꥢ̿)
//		fprintf(file,"Read SB %02X\n",ref_gb->get_regs()->SB);
		return ref_gb->get_regs()->SB;
	case 0xFF02://SC(ꥢ륳ȥ)
//		fprintf(file,"Read SC %02X\n",ref_gb->get_regs()->SC);
		return (ref_gb->get_regs()->SC&0x83)|0x7C;
	case 0xFF04://DIV(ǥХ?)
		return ref_gb->get_regs()->DIV;
	case 0xFF05://TIMA(ޥ)
		return ref_gb->get_regs()->TIMA;
	case 0xFF06://TMA(Ĵ)
		return ref_gb->get_regs()->TMA;
	case 0xFF07://TAC(ޥȥ)
		return ref_gb->get_regs()->TAC;
	case 0xFF0F://IF(ꤳߥե饰)
		return ref_gb->get_regs()->IF;
	case 0xFF40://LCDC(LCDȥ)
		return ref_gb->get_regs()->LCDC;
	case 0xFF41://STAT(LCDơ)
		return ref_gb->get_regs()->STAT|0x80;
	case 0xFF42://SCY(Y)
		return ref_gb->get_regs()->SCY;
	case 0xFF43://SCX(X)
		return ref_gb->get_regs()->SCX;
	case 0xFF44://LY(LCDC Yɸ)
		return ref_gb->get_regs()->LY;
	case 0xFF45://LYC(LY)
		return ref_gb->get_regs()->LYC;
	case 0xFF46://DMA(DMAž)
		return 0;
	case 0xFF47://BGP(طʥѥå)
		return ref_gb->get_regs()->BGP;
	case 0xFF48://OBP1(֥ȥѥå1)
		return ref_gb->get_regs()->OBP1;
	case 0xFF49://OBP2(֥ȥѥå2)
		return ref_gb->get_regs()->OBP2;
	case 0xFF4A://WY(ɥYɸ)
		return ref_gb->get_regs()->WY;
	case 0xFF4B://WX(ɥXɸ)
		return ref_gb->get_regs()->WX;

		//ʲ顼Ǥɲ
	case 0xFF4D://KEY1ƥ९åѹ
		return (speed?0x80:(ref_gb->get_cregs()->KEY1&1)?1:0x7E);
	case 0xFF4F://VBK(VRAMХڤؤ)
		return ref_gb->get_cregs()->VBK;
	case 0xFF51://HDMA1(ž)
		return dma_src>>8;
	case 0xFF52://HDMA2(ž)
		return dma_src&0xff;
	case 0xFF53://HDMA3(ž)
		return dma_dest>>8;
	case 0xFF54://HDMA4(ž貼)
		return dma_dest&0xff;
	case 0xFF55://HDMA5(ž¹)
		return (dma_executing?((dma_rest-1)&0x7f):0xFF);
	case 0xFF56://RP(ֳ)
		if (ref_gb->get_target()){
			if ((ref_gb->get_cregs()->RP&0xC0)==0xC0){
				dword *que=ref_gb->get_target()->get_cpu()->rp_que;
				int que_cnt=0;
				int cur;
				while((que[que_cnt]&0xffff)>rest_clock)	cur=que[que_cnt++]>>16;
//				fprintf(file,"read RP %02X\n",(ref_gb->get_cregs()->RP&1)|((cur&1)<<1)|0xC0);
				return (ref_gb->get_cregs()->RP&1)|((cur&1)<<1)|0xC0;

//				fprintf(file,"read RP %02X\n",(ref_gb->get_cregs()->RP&1)|((ref_gb->get_target()->get_cregs()->RP&1)<<1)|0xC0);
//				return (ref_gb->get_cregs()->RP&1)|((ref_gb->get_target()->get_cregs()->RP&1)<<1)|0xC0;
			}
			else{
//				fprintf(file,"read RP %02X\n",(ref_gb->get_cregs()->RP&1));
				return (ref_gb->get_cregs()->RP&1);
			}
		}
		else{
			if (ref_gb->hook_ext){ // եåޤ
				if ((ref_gb->get_cregs()->RP&0xC0)==0xC0)
					return (ref_gb->get_cregs()->RP&1)|(ref_gb->hook_proc.led()?2:0)|0xC0;
				else
					return (ref_gb->get_cregs()->RP&0xC1);
			}
			else
				return (ref_gb->get_cregs()->RP&0xC1);
		}
	case 0xFF68://BCPS(BGѥåȽ񤭹߻)
		return ref_gb->get_cregs()->BCPS;
	case 0xFF69://BCPD(BGѥåȽ񤭤ߥǡ)
		if (ref_gb->get_cregs()->BCPS&1)
			ret=ref_gb->get_lcd()->get_pal((ref_gb->get_cregs()->BCPS>>3)&7)[(ref_gb->get_cregs()->BCPS>>1)&3]>>8;
		else
			ret=ref_gb->get_lcd()->get_pal((ref_gb->get_cregs()->BCPS>>3)&7)[(ref_gb->get_cregs()->BCPS>>1)&3]&0xff;
/*		if (ref_gb->get_cregs()->BCPS&1)
			ret=ref_gb->get_renderer()->unmap_color(ref_gb->get_lcd()->get_pal((ref_gb->get_cregs()->BCPS>>3)&7)[(ref_gb->get_cregs()->BCPS>>1)&3])>>8;
		else
			ret=ref_gb->get_renderer()->unmap_color(ref_gb->get_lcd()->get_pal((ref_gb->get_cregs()->BCPS>>3)&7)[(ref_gb->get_cregs()->BCPS>>1)&3])&0xff;
*/		//ݥ󥿤ϥ󥯥Ȥʤ()
		return ret;
	case 0xFF6A://OCPS(OBJѥåȽ񤭤߻)
		return ref_gb->get_cregs()->OCPS;
	case 0xFF6B://OCPD(OBJѥåȽ񤭤ߥǡ)
		if (ref_gb->get_cregs()->OCPS&1)
			ret=ref_gb->get_lcd()->get_pal(((ref_gb->get_cregs()->OCPS>>3)&7)+8)[(ref_gb->get_cregs()->OCPS>>1)&3]>>8;
		else
			ret=ref_gb->get_lcd()->get_pal(((ref_gb->get_cregs()->OCPS>>3)&7)+8)[(ref_gb->get_cregs()->OCPS>>1)&3]&0xff;
/*		if (ref_gb->get_cregs()->OCPS&1)
			ret=ref_gb->get_renderer()->unmap_color(ref_gb->get_lcd()->get_pal(((ref_gb->get_cregs()->OCPS>>3)&7)+8)[(ref_gb->get_cregs()->OCPS>>1)&3])>>8;
		else
			ret=ref_gb->get_renderer()->unmap_color(ref_gb->get_lcd()->get_pal(((ref_gb->get_cregs()->OCPS>>3)&7)+8)[(ref_gb->get_cregs()->OCPS>>1)&3])&0xff;
*/		return ret;
	case 0xFF70://SVBK(RAMХڤؤ)
		return ref_gb->get_cregs()->SVBK;

	case 0xFFFF://IE(ꤳߥޥ)
		return ref_gb->get_regs()->IE;

	// undocumented register
	case 0xFF6C:
		return _ff6c&1;
	case 0xFF72:
		return _ff72;
	case 0xFF73:
		return _ff73;
	case 0xFF74:
		return _ff74;
	case 0xFF75:
		return _ff75&0x70;
	case 0xFF76:
		return 0;
	case 0xFF77:
		return 0;
	default:
		if (adr>0xFF0F&&adr<0xFF40){
			return ref_gb->get_apu()->read(adr);
		}
		else if ((adr>0xff70)&&(adr<0xff80))
			return ext_mem[adr-0xff71];
		else
			return 0;
	}
}

void cpu::io_write(word adr,byte dat)
{
		switch(adr){
		case 0xFF00://P1(ѥå)
			ref_gb->get_regs()->P1=dat;
			return;
		case 0xFF01://SB(ꥢ륷ꥢ̿)
			ref_gb->get_regs()->SB=dat;
			return;
		case 0xFF02://SC(ȥ)
			if (ref_gb->get_rom()->get_info()->gb_type==1){
				ref_gb->get_regs()->SC=dat&0x81;
				if ((dat&0x80)&&(dat&1)) // 
					seri_occer=total_clock+512;
			}
			else{ // GBCǤγĥ
				ref_gb->get_regs()->SC=dat&0x83;
				if ((dat&0x80)&&(dat&1)) // 
					if (dat&2)
						seri_occer=total_clock+512*8/32; // ž®̾32
					else
						seri_occer=total_clock+512*8;
			}
			return;
		case 0xFF04://DIV(ǥХ)
			ref_gb->get_regs()->DIV=0;
			return;
		case 0xFF05://TIMA(ޥ)
			ref_gb->get_regs()->TIMA=dat;
//			sys_clock=0;
			return;
		case 0xFF06://TMA(Ĵ)
			ref_gb->get_regs()->TMA=dat;
//			sys_clock=0;
			return;
		case 0xFF07://TAC(ޥȥ)
			if ((dat&0x04)&&!(ref_gb->get_regs()->TAC&0x04))
				sys_clock=0;
			ref_gb->get_regs()->TAC=dat;
			return;
		case 0xFF0F://IF(ꤳߥե饰)
			ref_gb->get_regs()->IF=dat;
			return;
		case 0xFF40://LCDC(LCDȥ)
			if ((dat&0x80)&&(!(ref_gb->get_regs()->LCDC&0x80))){
				ref_gb->get_regs()->LY=0;
				ref_gb->get_lcd()->clear_win_count();
			}
			ref_gb->get_regs()->LCDC=dat;
//			fprintf(file,"LCDC=%02X at line %d\n",dat,ref_gb->get_regs()->LY);
			return;
		case 0xFF41://STAT(LCDơ)
			if (ref_gb->get_rom()->get_info()->gb_type==1) // ꥸʥGBˤƤΤ褦ʸݤ餷
				if (!(ref_gb->get_regs()->STAT&0x02))
					ref_gb->get_regs()->IF|=INT_LCDC;

			ref_gb->get_regs()->STAT=(ref_gb->get_regs()->STAT&0x7)|(dat&0x78);
			return;
		case 0xFF42://SCY(Y)
			ref_gb->get_regs()->SCY=dat;
			return;
		case 0xFF43://SCX(X)
			ref_gb->get_regs()->SCX=dat;
			return;
		case 0xFF44://LY(LCDC Yɸ)
//			ref_gb->get_regs()->LY=0;
			ref_gb->get_lcd()->clear_win_count();
			return;
		case 0xFF45://LYC(LY)
			ref_gb->get_regs()->LYC=dat;
			return;
		case 0xFF46://DMA(DMAž)
			switch(dat>>5){
			case 0:
			case 1:
				memcpy(oam,ref_gb->get_rom()->get_rom()+dat*256,0xA0);
				break;
			case 2:
			case 3:
				memcpy(oam,ref_gb->get_mbc()->get_rom()+dat*256,0xA0);
				break;
			case 4:
				memcpy(oam,vram_bank+(dat&0x1F)*256,0xA0);
				break;
			case 5:
				memcpy(oam,ref_gb->get_mbc()->get_sram()+(dat&0x1F)*256,0xA0);
				break;
			case 6:
				if (dat&0x10)
					memcpy(oam,ram_bank+(dat&0x0F)*256,0xA0);
				else
					memcpy(oam,ram+(dat&0x0F)*256,0xA0);
				break;
			case 7:
				if (dat<0xF2){
					if (dat&0x10)
						memcpy(oam,ram_bank+(dat&0x0F)*256,0xA0);
					else
						memcpy(oam,ram+(dat&0x0F)*256,0xA0);
				}
				break;
			}
			return;
		case 0xFF47://BGP(طʥѥå)
			ref_gb->get_regs()->BGP=dat;
			return;
		case 0xFF48://OBP1(֥ȥѥå1)
			ref_gb->get_regs()->OBP1=dat;
			return;
		case 0xFF49://OBP2(֥ȥѥå2)
			ref_gb->get_regs()->OBP2=dat;
			return;
		case 0xFF4A://WY(ɥYɸ)
			ref_gb->get_regs()->WY=dat;
			return;
		case 0xFF4B://WX(ɥXɸ)
			ref_gb->get_regs()->WX=dat;
			return;

			//ʲ顼Ǥɲ
		case 0xFF4D://KEY1ƥ९åѹ
//			speed=dat&1;
			ref_gb->get_cregs()->KEY1=dat&1;
			speed_change=dat&1;
			return;
		case 0xFF4F://VBK(VRAMХڤؤ)
			if (dma_executing)
				return;
			vram_bank=vram+0x2000*(dat&0x01);
			ref_gb->get_cregs()->VBK=dat;//&0x01;
			return;
		case 0xFF51://HDMA1(ž)
			dma_src&=0x00F0;
			dma_src|=(dat<<8);
			return;
		case 0xFF52://HDMA2(ž)
			dma_src&=0xFF00;
			dma_src|=(dat&0xF0);
			return;
		case 0xFF53://HDMA3(ž)
			dma_dest&=0x00F0;
			dma_dest|=((dat&0xFF)<<8);
			return;
		case 0xFF54://HDMA4(ž貼)
			dma_dest&=0xFF00;
			dma_dest|=(dat&0xF0);
			return;
		case 0xFF55://HDMA5(ž¹)
			word tmp_adr;
			tmp_adr=0x8000+(dma_dest&0x1ff0);
//			fprintf(file,"%03d : %04X -> %04X  %d byte %s\n",ref_gb->get_regs()->LY,dma_src,dma_dest,((dat&0x7f)+1)*16,(dat&0x80)?"delay":"immidiately");
			if ((dma_src>=0x8000&&dma_src<0xA000)||(dma_src>=0xE000)||(!(tmp_adr>=0x8000&&tmp_adr<0xA000))){
				ref_gb->get_cregs()->HDMA5=0;
				return;
			}
			if (dat&0x80){ //HBlank
				if (dma_executing){
					dma_executing=false;
					dma_rest=0;
					ref_gb->get_cregs()->HDMA5=0xFF;
					return;
				}
				dma_executing=true;
				b_dma_first=true;
				dma_rest=(dat&0x7F)+1;
				ref_gb->get_cregs()->HDMA5=0;
/*				dma_dest_bank=vram_bank;
				if (dma_src<0x4000)
					dma_src_bank=ref_gb->get_rom()->get_rom();
				else if (dma_src<0x8000)
					dma_src_bank=ref_gb->get_mbc()->get_rom()-0x4000;
				else if (dma_src>=0xA000&&dma_src<0xC000)
					dma_src_bank=ref_gb->get_mbc()->get_sram()-0xA000;
				else if (dma_src>=0xC000&&dma_src<0xD000)
					dma_src_bank=ram-0xC000;
				else if (dma_src>=0xD000&&dma_src<0xE000)
					dma_src_bank=ram_bank-0xD000;
				else dma_src_bank=NULL;
*/			}
			else{ //̾DMA
				if (dma_executing){
					dma_executing=false;
					dma_rest=0;
					ref_gb->get_cregs()->HDMA5=0xFF;
//					fprintf(file,"dma stopped\n");
					return;
				}
				// ɤ顢HBlankʳʤ餤ĤǤOKߤ
//				if (!(((ref_gb->get_regs()->STAT&3)==1)||(!(ref_gb->get_regs()->LCDC&0x80)))){
//					ref_gb->get_cregs()->HDMA5=0;
//					return;
//				}

				dma_executing=false;
				dma_rest=0;
				ref_gb->get_cregs()->HDMA5=0xFF;

				switch(dma_src>>13){
				case 0:
				case 1:
					memcpy(vram_bank+(dma_dest&0x1ff0),ref_gb->get_rom()->get_rom()+(dma_src),16*(dat&0x7F)+16);
					break;
				case 2:
				case 3:
					memcpy(vram_bank+(dma_dest&0x1ff0),ref_gb->get_mbc()->get_rom()+(dma_src),16*(dat&0x7F)+16);
					break;
				case 4:
					break;
				case 5:
					memcpy(vram_bank+(dma_dest&0x1ff0),ref_gb->get_mbc()->get_sram()+(dma_src&0x1FFF),16*(dat&0x7F)+16);
					break;
				case 6:
					if (dma_src&0x1000)
						memcpy(vram_bank+(dma_dest&0x1ff0),ram_bank+(dma_src&0x0FFF),16*(dat&0x7F)+16);
					else
						memcpy(vram_bank+(dma_dest&0x1ff0),ram+(dma_src&0x0FFF),16*(dat&0x7F)+16);
					break;
				case 7:
					break;
				}
				dma_src+=((dat&0x7F)+1)*16;
				dma_dest+=((dat&0x7F)+1)*16;

				gdma_rest=456*2+((dat&0x7f)+1)*32*(speed?2:1); // CPU ѥ
			}
			return;
		case 0xFF56://RP(ֳ)
//			fprintf(file,"RP=%02X\n",dat);
			rp_que[que_cur++]=(((dword)dat)<<16)|((word)rest_clock);
			rp_que[que_cur]=0x00000000;
			ref_gb->get_cregs()->RP=dat;
			return;
		case 0xFF68://BCPS(BGѥåȽ񤭹߻)
			ref_gb->get_cregs()->BCPS=dat;
			return;
		case 0xFF69://BCPD(BGѥåȽ񤭤ߥǡ xBBBBBGG GGGRRRRR)
			if (ref_gb->get_cregs()->BCPS&1){
				ref_gb->get_lcd()->get_pal((ref_gb->get_cregs()->BCPS>>3)&7)[(ref_gb->get_cregs()->BCPS>>1)&3]=
				(ref_gb->get_lcd()->get_pal((ref_gb->get_cregs()->BCPS>>3)&7)[(ref_gb->get_cregs()->BCPS>>1)&3]&0xff)|(dat<<8);
			}
			else{
				ref_gb->get_lcd()->get_pal((ref_gb->get_cregs()->BCPS>>3)&7)[(ref_gb->get_cregs()->BCPS>>1)&3]=
				(ref_gb->get_lcd()->get_pal((ref_gb->get_cregs()->BCPS>>3)&7)[(ref_gb->get_cregs()->BCPS>>1)&3]&0xff00)|dat;
			}
			ref_gb->get_lcd()->get_mapped_pal((ref_gb->get_cregs()->BCPS>>3)&7)[(ref_gb->get_cregs()->BCPS>>1)&3]=
				ref_gb->get_renderer()->map_color(ref_gb->get_lcd()->get_pal((ref_gb->get_cregs()->BCPS>>3)&7)[(ref_gb->get_cregs()->BCPS>>1)&3]);
/*			if (ref_gb->get_cregs()->BCPS&1){
				ref_gb->get_lcd()->get_pal((ref_gb->get_cregs()->BCPS>>3)&7)[(ref_gb->get_cregs()->BCPS>>1)&3]=
					ref_gb->get_renderer()->map_color(((ref_gb->get_renderer()->unmap_color(ref_gb->get_lcd()->get_pal((ref_gb->get_cregs()->BCPS>>3)&7)[(ref_gb->get_cregs()->BCPS>>1)&3])&0xff)|(dat<<8)));
			}
			else{
				ref_gb->get_lcd()->get_pal((ref_gb->get_cregs()->BCPS>>3)&7)[(ref_gb->get_cregs()->BCPS>>1)&3]=
					ref_gb->get_renderer()->map_color(((ref_gb->get_renderer()->unmap_color(ref_gb->get_lcd()->get_pal((ref_gb->get_cregs()->BCPS>>3)&7)[(ref_gb->get_cregs()->BCPS>>1)&3])&0xff00)|(dat)));
			}*/
			ref_gb->get_cregs()->BCPD=dat;
			if (ref_gb->get_cregs()->BCPS&0x80)
				ref_gb->get_cregs()->BCPS=0x80|((ref_gb->get_cregs()->BCPS+1)&0x3f);
//			fprintf(file,"%d :BCPS = %02X\n",ref_gb->get_regs()->LY,dat);
			return;
		case 0xFF6A://OCPS(OBJѥåȽ񤭤߻)
			ref_gb->get_cregs()->OCPS=dat;
			return;
		case 0xFF6B://OCPD(OBJѥåȽ񤭤ߥǡ)
			if (ref_gb->get_cregs()->OCPS&1){
				ref_gb->get_lcd()->get_pal(((ref_gb->get_cregs()->OCPS>>3)&7)+8)[(ref_gb->get_cregs()->OCPS>>1)&3]=
				(ref_gb->get_lcd()->get_pal(((ref_gb->get_cregs()->OCPS>>3)&7)+8)[(ref_gb->get_cregs()->OCPS>>1)&3]&0xff)|(dat<<8);
			}
			else{
				ref_gb->get_lcd()->get_pal(((ref_gb->get_cregs()->OCPS>>3)&7)+8)[(ref_gb->get_cregs()->OCPS>>1)&3]=
				(ref_gb->get_lcd()->get_pal(((ref_gb->get_cregs()->OCPS>>3)&7)+8)[(ref_gb->get_cregs()->OCPS>>1)&3]&0xff00)|dat;
			}
			ref_gb->get_lcd()->get_mapped_pal(((ref_gb->get_cregs()->OCPS>>3)&7)+8)[(ref_gb->get_cregs()->OCPS>>1)&3]=
				ref_gb->get_renderer()->map_color(ref_gb->get_lcd()->get_pal(((ref_gb->get_cregs()->OCPS>>3)&7)+8)[(ref_gb->get_cregs()->OCPS>>1)&3]);
/*			if (ref_gb->get_cregs()->OCPS&1){
				ref_gb->get_lcd()->get_pal(((ref_gb->get_cregs()->OCPS>>3)&7)+8)[(ref_gb->get_cregs()->OCPS>>1)&3]=
					ref_gb->get_renderer()->map_color(((ref_gb->get_renderer()->unmap_color(ref_gb->get_lcd()->get_pal(((ref_gb->get_cregs()->OCPS>>3)&7)+8)[(ref_gb->get_cregs()->OCPS>>1)&3])&0xff)|(dat<<8)));
			}
			else{
				ref_gb->get_lcd()->get_pal(((ref_gb->get_cregs()->OCPS>>3)&7)+8)[(ref_gb->get_cregs()->OCPS>>1)&3]=
					ref_gb->get_renderer()->map_color(((ref_gb->get_renderer()->unmap_color(ref_gb->get_lcd()->get_pal(((ref_gb->get_cregs()->OCPS>>3)&7)+8)[(ref_gb->get_cregs()->OCPS>>1)&3])&0xff00)|(dat)));
			}*/
			ref_gb->get_cregs()->OCPD=dat;
			if (ref_gb->get_cregs()->OCPS&0x80)
				ref_gb->get_cregs()->OCPS=0x80|((ref_gb->get_cregs()->OCPS+1)&0x3f);
			return;
		case 0xFF70://SVBK(RAMХڤؤ)
//			if (dma_executing)
//				return;

			dat=(!(dat&7))?1:(dat&7);
			ref_gb->get_cregs()->SVBK=dat;
			ram_bank=ram+0x1000*dat;
			return;

		case 0xFFFF://IE(ꤳߥޥ)
			ref_gb->get_regs()->IE=dat;
//			ref_gb->get_regs()->IF=0;
//			fprintf(file,"IE = %02X\n",dat);
			return;

		// undocumented register
		case 0xFF6C:
			_ff6c=dat&1;
			return;
		case 0xFF72:
			_ff72=dat;
			return;
		case 0xFF73:
			_ff73=dat;
			return;
		case 0xFF74:
			_ff74=dat;
			return;
		case 0xff75:
			_ff75=dat&0x70;
			return;

		default:
			if (adr>0xFF0F&&adr<0xFF40){
				ref_gb->get_apu()->write(adr,dat,total_clock);
				return;
			}
			else if ((adr>0xff70)&&(adr<0xff80))
				ext_mem[adr-0xff71]=dat;
		}
}

void cpu::irq_process()
{
	if (int_desable){
		int_desable=false;
		return;
	}

	if ((ref_gb->get_regs()->IF&ref_gb->get_regs()->IE)&&(regs.I||halt)){//ꤳߤ
		if (halt)
			regs.PC++;
		write(regs.SP-2,regs.PC&0xFF);write(regs.SP-1,(regs.PC>>8));regs.SP-=2;
		if (ref_gb->get_regs()->IF&ref_gb->get_regs()->IE&INT_VBLANK){//VBlank
			regs.PC=0x40;
			ref_gb->get_regs()->IF&=0xFE;
			last_int=INT_VBLANK;
		}
		else if (ref_gb->get_regs()->IF&ref_gb->get_regs()->IE&INT_LCDC){//LCDC
			regs.PC=0x48;
			ref_gb->get_regs()->IF&=0xFD;
			last_int=INT_LCDC;
		}
		else if (ref_gb->get_regs()->IF&ref_gb->get_regs()->IE&INT_TIMER){//Timer
			regs.PC=0x50;
			ref_gb->get_regs()->IF&=0xFB;
			last_int=INT_TIMER;
		}
		else if (ref_gb->get_regs()->IF&ref_gb->get_regs()->IE&INT_SERIAL){//Serial
			regs.PC=0x58;
			ref_gb->get_regs()->IF&=0xF7;
			last_int=INT_SERIAL;
		}
		else if (ref_gb->get_regs()->IF&ref_gb->get_regs()->IE&INT_PAD){//Pad
			regs.PC=0x60;
			ref_gb->get_regs()->IF&=0xEF;
			last_int=INT_PAD;
		}
		else {}

		halt=false;
		regs.I=0;
//		ref_gb->get_regs()->IF=0;
	}
}

